/*
 *	IP NAT implementation by Michael Hasenstein
 *		mha@informatik.tu-chemnitz.de
 *		mha@debis-sfi.de
 *	(see RFC 1631)
 *
 *
 *      Copyright (c) 1997 by Michael Hasenstein
 *
 *      --------------------------------------------------------------------
 *	parts of this code, especially the packet matching stuff, have been
 *	taken from linux/net/ipv4/ip_fw.c
 *	see this file for copyrights for this code
 *      --------------------------------------------------------------------
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MODULE
#error This file must be compiled as a module.
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ip.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/icmp.h>
#include "ip_nat_module.h"

#ifdef KERNEL_2_0
static struct symbol_table ip_nat_module_syms = {
#include <linux/symtab_begin.h>
        X(register_ip_nat_app),
        X(unregister_ip_nat_app),
#include <linux/symtab_end.h>
};
#else  /* kernel 2.1 */
EXPORT_SYMBOL(register_ip_nat_app);
EXPORT_SYMBOL(unregister_ip_nat_app);
#endif


struct ip_nat *chain;       /* contains the nat-rules */


#ifdef MODULE
/*
 *  from linux/arch/i386/lib/checksum.c
 *  the procedure is not exported by the kernel,
 *  so I had to define it myself
 */
unsigned int csum_partial (const unsigned char * buff, int len,
			   unsigned int sum)
{
	__asm__("
	    testl $2, %%esi		# Check alignment.
	    jz 2f			# Jump if alignment is ok.
	    subl $2, %%ecx		# Alignment uses up two bytes.
	    jae 1f			# Jump if we had at least two bytes.
	    addl $2, %%ecx		# ecx was < 2.  Deal with it.
	    jmp 4f
1:	    movw (%%esi), %%bx
	    addl $2, %%esi
	    addw %%bx, %%ax
	    adcl $0, %%eax
2:
	    movl %%ecx, %%edx
	    shrl $5, %%ecx
	    jz 2f
	    testl %%esi, %%esi
1:	    movl (%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 4(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 8(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 12(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 16(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 20(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 24(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 28(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    lea 32(%%esi), %%esi
	    dec %%ecx
	    jne 1b
	    adcl $0, %%eax
2:	    movl %%edx, %%ecx
	    andl $0x1c, %%edx
	    je 4f
	    shrl $2, %%edx
	    testl %%esi, %%esi
3:	    adcl (%%esi), %%eax
	    lea 4(%%esi), %%esi
	    dec %%edx
	    jne 3b
	    adcl $0, %%eax
4:	    andl $3, %%ecx
	    jz 7f
	    cmpl $2, %%ecx
	    jb 5f
	    movw (%%esi),%%cx
	    leal 2(%%esi),%%esi
	    je 6f
	    shll $16,%%ecx
5:	    movb (%%esi),%%cl
6:	    addl %%ecx,%%eax
	    adcl $0, %%eax
7:	    "
	: "=a"(sum)
	: "0"(sum), "c"(len), "S"(buff)
	: "bx", "cx", "dx", "si");
	return(sum);
}
#endif  /* MODULE */

#ifdef CONFIG_PROC_FS
static int ip_nat_procinfo (char *buffer, char **start, off_t offset,
                            int length, int reset)
{
	off_t pos=0, begin=0;
        struct ip_nat *i;
        unsigned long flags;
        int len, p;
        int last_len = 0;
	
	i = chain;

	len=sprintf(buffer, "IP NAT rules\n");

	save_flags(flags);
        cli();

        while(i!=NULL)
        {
                len+=sprintf(buffer+len,
			"%X %X %08lX/%08lX->%08lX/%08lX ",
			i->id, i->protocol,
                	ntohl(i->src.s_addr), ntohl(i->src_msk.s_addr),
                	ntohl(i->dst.s_addr), ntohl(i->dst_msk.s_addr));
                len+=sprintf(buffer+len,
			"%08lX/%08lX %08lX/%08lX %.16s %08lX %X ",
                	ntohl(i->src_nat.s_addr), ntohl(i->src_nat_msk.s_addr),
                	ntohl(i->dst_nat.s_addr), ntohl(i->dst_nat_msk.s_addr),
                	(i->via_dev) ? i->via_dev->name : "-",
                	ntohl(i->via.s_addr), i->flag);
                len+=sprintf(buffer+len,
			"%u %u %-9lu %-9lu",
			i->nsp,i->ndp, i->pcnt,i->bcnt);
                for (p = 0; p < IP_NAT_MAX_PORTS; p++)
                        len+=sprintf(buffer+len, " %u", i->ports[p]);
                buffer[len++]='\n';
                buffer[len]='\0';
                pos=begin+len;
                if(pos<offset)
                {
                        len=0;
                        begin=pos;
                }
                else if(pos>offset+length)
                {
                        len = last_len;
                        break;          
                }
                else if(reset)
                {
                  /* This needs to be done at this specific place! */
                        i->pcnt=0L;
                        i->bcnt=0L;
                }
                last_len = len;
                i=i->next;
        }
        restore_flags(flags);
        *start=buffer+(offset-begin);
        len-=(offset-begin);
        if(len>length)
                len=length;     
	
        return(len);
}
#endif  /* CONFIG_PROC_FS */


/*
 *      Returns 1 if the port is matched by the vector, 0 otherwise
 */

int port_match (unsigned short *portptr,int nports,unsigned short port,
                int range_flag)
{
        if (!nports)
                return 1;
        if (range_flag) {
                if (portptr[0] <= port && port <= portptr[1])
                        return(1);
                nports -= 2;
                portptr += 2;
        }
        while (nports-- > 0) {
                if (*portptr++ == port) 
                        return(1);
	}
        return(0);
}

void flush_rules (void) {
	struct ip_nat *s,*t;
	unsigned long flags;

	save_flags(flags);
	cli();
	
	if (chain)
		MOD_DEC_USE_COUNT;

	t = chain;
	while (t) {
		s = t->next;
		kfree_s(t,sizeof(*t));
		t = s;
	}
	chain = NULL;
	
	restore_flags(flags);
	return;
}

int delete_rule (struct ip_nat_set *r)
{
	struct ip_nat *s,*t;
	unsigned long flags;

	save_flags(flags);
	cli();

	r->timeout = 0;  /* flag 'found rule' */
	t = chain;
	s = NULL;  /* pointer to predecessor */
	while (t) {
		if (t->id == r->nat.id) {
			if (s) {
				s->next = t->next;
				kfree_s(t, sizeof(*t));
			} else {
				/* first rule in chain */
				chain = t->next;
				kfree_s(t, sizeof(*t));
			}
			r->timeout = 1;
		}
		s = t;
		t = t->next;
	}

	restore_flags(flags);

	if (r->timeout) {
		if (!chain)
			/* no rules left, so module is unused */
			MOD_DEC_USE_COUNT;
		return(0);
	}
	else  /* not found */
		return(EINVAL);
}

int insert_rule (struct ip_nat_set *r) {

	struct ip_nat *s,*t;
	struct ip_nat *last;
	struct ip_nat *new;
	unsigned long flags;
	
	save_flags(flags);
	
	new = kmalloc(sizeof(struct ip_nat), GFP_ATOMIC);
	if (new == NULL) {
                printk("ip-nat-module: insert_rule:  malloc said no\n");
                return(ENOMEM);
        }
        memcpy(new, &(r->nat), sizeof(struct ip_nat));
	
	new->pcnt = 0;
	new->bcnt = 0;
	new->next = NULL;

	if ((r->vianame)[0])
                if (!(new->via_dev = dev_get(r->vianame)))
                        new->via_dev = NULL;

	if ((r->nat.ndp + r->nat.nsp) > IP_NAT_MAX_PORTS)
		return(EINVAL);

	/* just to be sure... */
	new->src.s_addr     &= new->src_msk.s_addr;
	new->dst.s_addr     &= new->dst_msk.s_addr;
	new->src_nat.s_addr &= new->src_nat_msk.s_addr;
	new->dst_nat.s_addr &= new->dst_nat_msk.s_addr;
		
	cli();
	
	if (!chain) {
		chain = new;
		if (!new->id)
			new->id = 1;
		restore_flags(flags);
		MOD_INC_USE_COUNT;
		return(0);
	}

		
	t = chain;
	last = NULL;  /* pointer to predecessor */
	while(t) {
		if ((t->id >= new->id) && (new->id)) {
			new->next = t;
			if (last)
				last->next = new;
			else
				chain = new;
			if (t->id == new->id) {
				s = t;
				s->id++;
				while ((s) && (s->id == s->next->id)) {
					s = s->next;
					s->id++;
				}
			}
			restore_flags(flags);
			return(0);
		}
		last = t;
		t = t->next;
	}

	if (!new->id)
		new->id = last->id + 1;
	last->next = new;

	restore_flags(flags);
	return(0);
}

void print_packet (struct iphdr *ip, struct device *dev, struct ip_nat *t)
{
	__u32 *opt = (__u32 *) (ip + 1);
	__u16 src_port=0xFFFF, dst_port=0xFFFF;
	int opti;

        struct tcphdr  *tcp =(struct tcphdr *) ((unsigned long *)ip+ip->ihl);
        struct udphdr  *udp =(struct udphdr *) ((unsigned long *)ip+ip->ihl);
        struct icmphdr *icmp=(struct icmphdr *)((unsigned long *)ip+ip->ihl);

	printk(dev ? dev->name : "-");
	switch(ip->protocol)
	{
		case IPPROTO_TCP:
			printk(" TCP ");
			src_port = ntohs(tcp->source);
			dst_port = ntohs(tcp->dest);
			break;
		case IPPROTO_UDP:
			printk(" UDP ");
			src_port = ntohs(udp->source);
			dst_port = ntohs(udp->dest);
			break;
		case IPPROTO_ICMP:
			printk(" ICMP/%d ", icmp->type);
			break;
		default:
			printk(" PROTO=%d ", ip->protocol);
			break;
	}
	printk("src:");
	print_ip(ip->saddr);
	if(ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP)
		printk(":%hu", src_port);
	printk(" dst:");
	print_ip(ip->daddr);
	if(ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP)
		printk(":%hu", dst_port);
	printk(" nat-src:");
	print_ip(t->src_nat.s_addr);
	printk(" nat-dst:");
	print_ip(t->dst_nat.s_addr);
	printk(" L=%hu S=0x%2.2hX I=%hu F=0x%4.4hX T=%hu",
		ntohs(ip->tot_len), ip->tos, ntohs(ip->id),
		ip->frag_off, ip->ttl);
	for (opti=0; 
	     opti<(ip->ihl - sizeof(struct iphdr)/4);
	     opti++)
	printk(" O=0x%8.8X", *opt++);
}

int ip_nat_mod_cntl (int optname, void *tmp_nat, int optlen)
{
	struct ip_nat_set *tmp = (struct ip_nat_set*) tmp_nat;

	if (optlen != sizeof(struct ip_nat_set))
		return(EINVAL);

	switch (tmp->command)
	{
		case IP_NAT_CMD_FLUSH:
			flush_rules();
			break;
		case IP_NAT_CMD_TIMEOUTS:
			break;
		case IP_NAT_CMD_DELETE:
			return (delete_rule(tmp));
			break;
		case IP_NAT_CMD_INSERT:
			return (insert_rule(tmp));
			break;
		default:
			printk("ip_nat_mod_ctl: unknown request\n");
			return(EINVAL);
	}
	
		
	return(0);
}

/*
 *  static address translation:
 *  rewrite source- and/or destination ip in the packet
 *  to the new source- and/or destination ip
 */

void ip_nat_translate (struct sk_buff *skb, struct iphdr *ip,
		       struct tcphdr *tcp, struct udphdr *udp,
		       __u32 saddr, __u32 s_msk,
		       __u32 daddr, __u32 d_msk,
		       __u16 sport, __u16 dport)
{
	int len = ntohs(ip->tot_len)-ip->ihl*4;
	
	if (saddr)
		skb->saddr = ip->saddr = (saddr&s_msk) | (ip->saddr&(~s_msk));
	
	if (daddr)
		skb->daddr = ip->daddr = (daddr&d_msk) | (ip->daddr&(~d_msk));

	if (sport) {
		if (ip->protocol == IPPROTO_TCP)
			tcp->source = sport;
		else if (ip->protocol == IPPROTO_UDP)
			udp->source = sport;
	}

	if (dport) {
		if (ip->protocol == IPPROTO_TCP)
			tcp->dest = dport;
		else if (ip->protocol == IPPROTO_UDP)
			udp->dest = dport;
	}

	/*
	 *  that was it, but now we have to recompute
	 *  the checksums
	 */

	switch (ip->protocol) {
		case IPPROTO_TCP:
			tcp->check = 0;
			tcp->check = csum_tcpudp_magic(ip->saddr,
					ip->daddr,len,IPPROTO_TCP,
					csum_partial((char *)tcp,len,0));
			if (tcp->check==0)
				tcp->check = 0xFFFF;
			break;
		case IPPROTO_UDP:
			udp->check = 0;
			udp->check = csum_tcpudp_magic(ip->saddr,
					ip->daddr,len,IPPROTO_UDP,
					csum_partial((char *)udp,len,0));
			if (udp->check==0)
				udp->check = 0xFFFF;
			break;
		default:
			break;
	}
	ip_send_check(ip);

	return;
}

void ip_nat_mod_fnct (u_char direction, struct sk_buff *skb)
{
	struct ip_nat *t;
	struct iphdr *ip;
        struct tcphdr  *tcp;
        struct udphdr  *udp;
        struct icmphdr *icmp;
	__u32 src,dst;
	__u16 src_port=0xFFFF, dst_port=0xFFFF, icmp_type=0xFF;
	char notcpsyn=0, notcpack=0, reverse;
	unsigned short offset;

#ifdef KERNEL_2_0
	if (direction == IP_NAT_IN)
		ip   = skb->h.iph;
	else  /* IP_NAT_OUT */
		ip   = skb->ip_hdr;
#else  /* kernel 2.1 */
	ip   = skb->nh.iph;
#endif
        tcp  = (struct tcphdr *) ((unsigned long *)ip+ip->ihl);
        udp  = (struct udphdr *) ((unsigned long *)ip+ip->ihl);
        icmp = (struct icmphdr *)((unsigned long *)ip+ip->ihl);

	if (!chain)
		return;

	if (ip->version != 4)
		return;

	src = ip->saddr;
	dst = ip->daddr;
	offset = ntohs(ip->frag_off) & IP_OFFSET;

	switch(ip->protocol) 
        {
                case IPPROTO_TCP:
                  /* ports stay 0xFFFF if it is not the first fragment */
                        if (!offset) {
                                src_port=ntohs(tcp->source);
                                dst_port=ntohs(tcp->dest);
                                if(!tcp->ack && !tcp->rst)
                                        notcpack=1;
                                if(!tcp->syn || !notcpack)
                                        notcpsyn=1;
                        }
                        break;
                case IPPROTO_UDP:
                  /* ports stay 0xFFFF if it is not the first fragment */
                        if (!offset) {
                                src_port=ntohs(udp->source);
                                dst_port=ntohs(udp->dest);
                        }
                        break;
                case IPPROTO_ICMP:
                  /* icmp_type stays 255 if it is not the first fragment */
                        if (!offset)
                                icmp_type=(__u16)(icmp->type);
                        break;
                default:
                        break;
        }

	for (t=chain; t; t=t->next) 
	{
		/*
		 *  match direction (or option 'bidirectional' is set)
		 */

		if (direction & (t->flag&IP_NAT_DIRECTION_MASK))
			reverse = 0;
		else
			reverse = 1;

		if (reverse && !(t->flag&IP_NAT_BIDIR))
			continue;
		
		/*
		 *  match source and destination ip's
		 */

		if (reverse) {
			/* this is more complicated... */
			if (t->dst_nat.s_addr) {
				if ((src&t->dst_nat_msk.s_addr) !=
				     t->dst_nat.s_addr)
					continue;
			} else {
				if ((src&t->dst_msk.s_addr) != t->dst.s_addr)
					continue;
			}
			if (t->src_nat.s_addr) {
				if ((dst&t->src_nat_msk.s_addr) !=
				     t->src_nat.s_addr)
					continue;
			} else {
				if ((dst&t->src_msk.s_addr) != t->src.s_addr)
					continue;
			}
		} else {
			/* ...than this case */
			if (!((src&t->src_msk.s_addr)==t->src.s_addr &&
			      (dst&t->dst_msk.s_addr)==t->dst.s_addr))
				continue;
		}

                /*
                 *  look for a protocol match
                 */
		
		if ((t->protocol) && (t->protocol != ip->protocol))
			continue;

                /*
                 *  look for a VIA address match 
                 */
		
                if ((t->via.s_addr) && (t->via.s_addr != skb->dev->pa_addr))
			continue;

                /*
                 *  look for a VIA device match 
                 */
		
                if ((t->via_dev) && (t->via_dev != skb->dev))
			continue;
	
		/*
		 *  match ports
		 */

		if (t->nsp + t->ndp) {
			if (offset)
				printk("IP NAT-Warning: found fragment, ignoring ports\n");
			if((ip->protocol==IPPROTO_ICMP &&
				!port_match(&t->ports[0], t->nsp,
					icmp_type,t->flag&IP_NAT_SRNG)) ||
				!(ip->protocol==IPPROTO_ICMP || ((!reverse) &&
				port_match(&t->ports[0], t->nsp, src_port,
					t->flag&IP_NAT_SRNG) &&
				port_match(&t->ports[t->nsp], t->ndp, dst_port,
					t->flag&IP_NAT_DRNG)) || ((reverse) &&
				port_match(&t->ports[0], t->nsp, dst_port,
					t->flag&IP_NAT_SRNG) &&
				port_match(&t->ports[t->nsp], t->ndp, src_port,
					t->flag&IP_NAT_DRNG))))
			{
				continue;
			}
		}


		/* if we get here, we've found a matching rule */

		/*
		 *  print matching packets
		 */

		if (t->flag & IP_NAT_LOG) {
			if (direction == IP_NAT_IN)
				printk("IP NAT in (before) ");
			else
				printk("IP NAT out (before) ");
			print_packet(ip, skb->dev, t);
			printk("\n");
		}

		/*
		 *  process packet according to rule
		 */

		t->pcnt++;
		t->bcnt+=ntohs(ip->tot_len);

		/* do-nothing rule */
		if (t->flag & IP_NAT_SKIP)
			return;
		
		if (t->flag & IP_NAT_DYNAMIC) {
			/* not implemented yet */
		} else {
			/* static address translation */
			if (reverse) {
				/* reverse match (bidirectional rule) */
				ip_nat_translate(skb, ip, tcp, udp,
						 t->dst.s_addr,
						 t->dst_msk.s_addr,
					         t->src.s_addr,
						 t->src_msk.s_addr,
						 t->dst_nat_port,
						 t->src_nat_port);
			} else {
				/* direct match */
				ip_nat_translate(skb, ip, tcp, udp,
						 t->src_nat.s_addr,
						 t->src_nat_msk.s_addr,
					         t->dst_nat.s_addr,
						 t->dst_nat_msk.s_addr,
						 t->src_nat_port,
						 t->dst_nat_port);
			}

			/* change route-address only if target host is local */
			if (dst == skb->raddr)
				skb->raddr = ip->daddr;
		}
		
		/*
		 *  print matching packets
		 */

		if (t->flag & IP_NAT_LOG) {
			if (direction == IP_NAT_IN)
				printk("IP NAT in (after) ");
			else
				printk("IP NAT out (after) ");
			print_packet(ip, skb->dev, t);
			printk("\n");
		}


		return;
		
	}  /* loop */

	return;
}

int register_ip_nat_app (void)
{
	return(0);
}

int unregister_ip_nat_app (void)
{
	return(0);
}



int init_module (void)
{
#ifdef KERNEL_2_0
	register_symtab (&ip_nat_module_syms);
#endif

	/* tell the kernel the addresses of these two functions */
        if ( register_nat(ip_nat_mod_fnct, ip_nat_mod_cntl) != 0 )
                return(-EIO);

#ifdef CONFIG_PROC_FS
        proc_net_register(&(struct proc_dir_entry) {
                PROC_NET_IPNAT, 6, "ip_nat",
                S_IFREG | S_IRUGO, 1, 0, 0,
                0, &proc_net_inode_operations,
                ip_nat_procinfo
        });
#endif

        return(0);
}

void cleanup_module (void)
{
#ifdef CONFIG_PROC_FS
	proc_net_unregister(PROC_NET_IPNAT);
#endif
	flush_rules();

	unregister_nat();
}
