/*
 *	IP NAT implementation by Michael Hasenstein
 *		mha@informatik.tu-chemnitz.de
 *		mha@debis-sfi.de
 *	(see RFC 1631)
 *
 *
 *      Copyright (c) 1997 by Michael Hasenstein
 *
 *      --------------------------------------------------------------------
 *	parts of this code, especially the packet matching stuff, have been
 *	taken from linux/net/ipv4/ip_fw.c
 *	see this file for copyrights for this code
 *      --------------------------------------------------------------------
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MODULE
#error This file must be compiled as a module.
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ip.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/icmp.h>
#include "ip_nat_module.h"

#ifdef KERNEL_2_0
static struct symbol_table ip_nat_module_syms = {
#include <linux/symtab_begin.h>
        X(register_ip_nat_app),
        X(unregister_ip_nat_app),
#include <linux/symtab_end.h>
};
#else  /* kernel 2.1 */
EXPORT_SYMBOL(register_ip_nat_app);
EXPORT_SYMBOL(unregister_ip_nat_app);
#endif


struct ip_nat *chain;       /* contains the nat-rules */


#ifdef MODULE
/*
 *  from linux/arch/i386/lib/checksum.c
 *  the procedure is not exported by the kernel,
 *  so I had to define it myself
 */
unsigned int csum_partial (const unsigned char * buff, int len,
			   unsigned int sum)
{
	__asm__("
	    testl $2, %%esi		# Check alignment.
	    jz 2f			# Jump if alignment is ok.
	    subl $2, %%ecx		# Alignment uses up two bytes.
	    jae 1f			# Jump if we had at least two bytes.
	    addl $2, %%ecx		# ecx was < 2.  Deal with it.
	    jmp 4f
1:	    movw (%%esi), %%bx
	    addl $2, %%esi
	    addw %%bx, %%ax
	    adcl $0, %%eax
2:
	    movl %%ecx, %%edx
	    shrl $5, %%ecx
	    jz 2f
	    testl %%esi, %%esi
1:	    movl (%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 4(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 8(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 12(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 16(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 20(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 24(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    movl 28(%%esi), %%ebx
	    adcl %%ebx, %%eax
	    lea 32(%%esi), %%esi
	    dec %%ecx
	    jne 1b
	    adcl $0, %%eax
2:	    movl %%edx, %%ecx
	    andl $0x1c, %%edx
	    je 4f
	    shrl $2, %%edx
	    testl %%esi, %%esi
3:	    adcl (%%esi), %%eax
	    lea 4(%%esi), %%esi
	    dec %%edx
	    jne 3b
	    adcl $0, %%eax
4:	    andl $3, %%ecx
	    jz 7f
	    cmpl $2, %%ecx
	    jb 5f
	    movw (%%esi),%%cx
	    leal 2(%%esi),%%esi
	    je 6f
	    shll $16,%%ecx
5:	    movb (%%esi),%%cl
6:	    addl %%ecx,%%eax
	    adcl $0, %%eax
7:	    "
	: "=a"(sum)
	: "0"(sum), "c"(len), "S"(buff)
	: "bx", "cx", "dx", "si");
	return(sum);
}
#endif  /* MODULE */

#ifdef CONFIG_PROC_FS
static int ip_nat_procinfo (char *buffer, char **start, off_t offset,
                            int length, int reset)
{
	off_t pos=0, begin=0;
        struct ip_nat *i;
        unsigned long flags;
        int len, p;
        int last_len = 0;
	
	i = chain;

	len=sprintf(buffer, "IP NAT rules\n");
	len+=sprintf(buffer+len,"id prot     src/msk    ->    dst/msk         src-new/msk:port    dst-new/msk:port  dev   ip    flag #s-ports #d-ports timeout  packets    bytes       ports(src+dst)\n");

	save_flags(flags);
        cli();

        while(i) {
                len+=sprintf(buffer+len,
			"%X   %X  %08lX/%08lX->%08lX/%08lX ",
			i->id, i->protocol,
                	ntohl(i->src), ntohl(i->src_msk),
                	ntohl(i->dst), ntohl(i->dst_msk));
                len+=sprintf(buffer+len,
			"%08lX/%08lX:%u %08lX/%08lX:%u %.16s %08lX %X     ",
                	ntohl(i->src_nat), ntohl(i->src_nat_msk), i->src_nat_port,
                	ntohl(i->dst_nat), ntohl(i->dst_nat_msk), i->dst_nat_port,
                	(i->via_dev) ? i->via_dev->name : "-",
                	ntohl(i->via), i->flag);
                len+=sprintf(buffer+len,
			"%u      %u      %u   %-9lu %-9lu",
			i->nsp,i->ndp,i->timeout,i->pcnt,i->bcnt);
                for (p = 0; p < IP_NAT_MAX_PORTS; p++)
                        len+=sprintf(buffer+len, " %u", i->ports[p]);
                buffer[len++]='\n';
                buffer[len]='\0';
                pos=begin+len;
                if(pos<offset)
                {
                        len=0;
                        begin=pos;
                }
                else if(pos>offset+length)
                {
                        len = last_len;
                        break;          
                }
                last_len = len;
                i=i->next;
        }
        restore_flags(flags);

        *start=buffer+(offset-begin);
        len-=(offset-begin);
        if(len>length)
                len=length;     
	
        return(len);
}
#ifdef CONFIG_VIRTUAL_SERVER
static int ip_nat_vs_procinfo (char *buffer, char **start, off_t offset,
                            int length, int reset)
{
	off_t pos=0, begin=0;
        struct ip_nat *i;
        unsigned long flags;
        int len;
        int last_len = 0;
	struct server *s;
	struct client *c;

	i = chain;

	len=sprintf(buffer, "IP NAT virtual server rules\n");

	save_flags(flags);
        cli();

        while(i) {
		if (i->flag & IP_NAT_VIRT_SERVER) {
			/* print all servers */
			for (s=i->dynamic.vs->sr; s; s=s->next) {
				len+=sprintf(buffer+len, "S %u: %08lX:%X %-9lu %-9lu %u %u",
					i->id, ntohl(s->ip), s->port, s->pcnt, s->bcnt, 
					s->connections, s->weight);
				if (i->dynamic.vs->last == s)
					len+=sprintf(buffer+len," L");
				buffer[len++]='\n';
			}
			/* print all clients */
			for (c=i->dynamic.vs->cl; c; c=c->next) {
				len+=sprintf(buffer+len, "C %u: %08lX:%X %u", 
					     i->id, ntohl(c->ip), c->port, c->protocol);
				buffer[len++]='\n';
			}
			buffer[len]='\0';
			pos=begin+len;
			if(pos<offset)
			{
				len=0;
				begin=pos;
			}
			else if(pos>offset+length)
			{
				len = last_len;
				break;          
			}
			last_len = len;
		}
                i=i->next;
	}
        restore_flags(flags);

        *start=buffer+(offset-begin);
        len-=(offset-begin);
        if(len>length)
                len=length;     

        return(len);
}
#endif  /* CONFIG_VIRTUAL_SERVER */
#endif  /* CONFIG_PROC_FS */


/*
 *      Returns 1 if the port is matched by the vector, 0 otherwise
 */

int port_match (unsigned short *portptr,int nports,unsigned short port,
                int range_flag)
{
        if (!nports)
                return 1;
        if (range_flag) {
                if (portptr[0] <= port && port <= portptr[1])
                        return(1);
                nports -= 2;
                portptr += 2;
        }
        while (nports-- > 0) {
                if (*portptr++ == port) 
                        return(1);
	}
        return(0);
}

void flush_rules (void) {
	struct ip_nat *s,*t;
	unsigned long flags;
#ifdef CONFIG_VIRTUAL_SERVER
	struct server *vs_s1, *vs_s2;
	struct client *vs_c1, *vs_c2;
#endif
#ifdef CONFIG_DYNAMIC_NAT
#endif

	if (chain)
		MOD_DEC_USE_COUNT;

	t = chain;

	save_flags(flags);
	cli();
	chain = NULL;
	restore_flags(flags);

	while (t) {
		s = t->next;
#ifdef CONFIG_VIRTUAL_SERVER
		if (t->flag & IP_NAT_VIRT_SERVER) {
			/* delete all servers & clients */
			vs_s1 = t->dynamic.vs->sr;
			vs_c1 = t->dynamic.vs->cl;
			save_flags(flags);
			cli();
			t->dynamic.vs->sr = NULL;
			t->dynamic.vs->cl = NULL;
			restore_flags(flags);
			while (vs_c1) {
				vs_c2 = vs_c1->next;
				del_timer(&(vs_c1->timer));
				kfree_s(vs_c1,sizeof(*vs_c1));
				vs_c1 = vs_c2;
			}
			while (vs_s1) {
				vs_s2 = vs_s1->next;
				kfree_s(vs_s1,sizeof(*vs_s1));
				vs_s1 = vs_s2;
			}
			/* delete virtual-server struct */
			kfree_s(t->dynamic.vs,sizeof(*(t->dynamic.vs)));
		}
#endif
#ifdef CONFIG_DYNAMIC_NAT
		if (t->flag & IP_NAT_DYNAMIC) {
		}
#endif
		kfree_s(t,sizeof(*t));
		t = s;
	}
	
	return;
}

#ifdef CONFIG_VIRTUAL_SERVER
int delete_vs_rule (struct ip_nat *t)
{
	unsigned long flags;
	struct server *vs_s1, *vs_s2;
	struct client *vs_c1, *vs_c2;

	/* delete all servers & clients */
	vs_s1 = t->dynamic.vs->sr;
	vs_c1 = t->dynamic.vs->cl;
	save_flags(flags);
	cli();
	t->dynamic.vs->sr = NULL;
	t->dynamic.vs->cl = NULL;
	t->dynamic.vs->last = NULL;
	restore_flags(flags);
	while (vs_c1) {
		vs_c2 = vs_c1->next;
		del_timer(&(vs_c1->timer));
		kfree_s(vs_c1,sizeof(*vs_c1));
		vs_c1 = vs_c2;
	}
	while (vs_s1) {
		vs_s2 = vs_s1->next;
		kfree_s(vs_s1,sizeof(*vs_s1));
		vs_s1 = vs_s2;
	}
	/* delete virtual-server struct */
	kfree_s(t->dynamic.vs,sizeof(*(t->dynamic.vs)));

	return (1);
}

int delete_vs_server (struct ip_nat *t, struct ip_nat_set *r)
{
	int found = 0;
	unsigned long flags;
	struct server *s, *last_s;
	struct client *c, *last_c;

	s = t->dynamic.vs->sr;
	last_s = NULL;
	while (s) {
		if (s->ip == r->nat.dst) {
			found = 1;
			/* delete all clients for this server */
			c = t->dynamic.vs->cl;
			last_c = NULL;
			while (c) {
				if (c->server->ip == t->dst) {
					save_flags(flags);
					cli();
					if (last_c)
						last_c->next = c->next;
					else
						t->dynamic.vs->cl = c->next;
					restore_flags(flags);
					del_timer(&(c->timer));
					kfree_s(c,sizeof(*c));
				}
				last_c = c;
				c = c->next;
			}
			/* delete this server */
			save_flags(flags);
			cli();
			if (last_s)
				last_s->next = s->next;
			else
				t->dynamic.vs->sr = s->next;
			if (t->dynamic.vs->last->ip == r->nat.dst)
				t->dynamic.vs->last=NULL;
			restore_flags(flags);
		}
		last_s = s;
		s = s->next;
	}

	return(found);
}
#endif

#ifdef CONFIG_DYNAMIC_NAT
int delete_dyn_rule (struct ip_nat *t, struct ip_nat_set *r)
{
	int found = 0;

	return(found);
}
#endif

int delete_rule (struct ip_nat_set *r)
{
	struct ip_nat *s,*t;
	unsigned long flags;
	int found = 0;

	t = chain;
	s = NULL;  /* pointer to predecessor */
	while (t && !found) {
		if (t->id == r->nat.id) {
			if (s) {
#ifdef CONFIG_VIRTUAL_SERVER
				if ((t->flag & IP_NAT_VIRT_SERVER) && (r->nat.dst)) {
					if (!delete_vs_server(t,r))
						return(EINVAL);
					else
						return(0);
				} else if (t->flag & IP_NAT_VIRT_SERVER)
					found = delete_vs_rule(t);
				else {
#endif
#ifdef CONFIG_DYNAMIC_NAT
#endif
					save_flags(flags);
					cli();
					s->next = t->next;
					restore_flags(flags);
					found = 1;
#if defined(CONFIG_VIRTUAL_SERVER) || defined(CONFIG_DYNAMIC_NAT)
				}
#endif
			} else {
				/* first rule in chain */
#ifdef CONFIG_VIRTUAL_SERVER
				if ((t->flag & IP_NAT_VIRT_SERVER) && (r->nat.dst)) {
					if (!delete_vs_server(t,r))
						return(EINVAL);
					else
						return(0);
				} else if (t->flag & IP_NAT_VIRT_SERVER) {
					found = delete_vs_rule(t);
					save_flags(flags);
					cli();
					chain = t->next;
					restore_flags(flags);
				} else {
#endif
#ifdef CONFIG_DYNAMIC_NAT
#endif
					save_flags(flags);
					cli();
					chain = t->next;
					restore_flags(flags);
					found = 1;
#if defined(CONFIG_VIRTUAL_SERVER) || defined(CONFIG_DYNAMIC_NAT)
				}
#endif
			}
		}
		s = t;
		t = t->next;
	}

	if (found) {
		kfree_s(t, sizeof(*t));
		if (!chain)
			/* no rules left, so module is unused */
			MOD_DEC_USE_COUNT;
		return(0);
	} else
		return(EINVAL);
}

int insert_rule (struct ip_nat_set *r) {

	struct ip_nat *s,*t;
	struct ip_nat *last;
	struct ip_nat *new;
	unsigned long flags;
	
	save_flags(flags);
	
	new = kmalloc(sizeof(struct ip_nat), GFP_ATOMIC);
	if (new == NULL) {
                printk("ip-nat-module: insert_rule:  malloc said no\n");
                return(ENOMEM);
        }
        memcpy(new, &(r->nat), sizeof(struct ip_nat));
	
	new->pcnt = 0;
	new->bcnt = 0;
	new->next = NULL;

	if ((r->vianame)[0])
                if (!(new->via_dev = dev_get(r->vianame)))
                        new->via_dev = NULL;

	if ((r->nat.ndp + r->nat.nsp) > IP_NAT_MAX_PORTS)
		return(EINVAL);

	/* just to make sure... */
	new->src     &= new->src_msk;
	new->dst     &= new->dst_msk;
	new->src_nat &= new->src_nat_msk;
	new->dst_nat &= new->dst_nat_msk;

	/* dynamic rules are always bidirectional */
	if ((new->flag & IP_NAT_VIRT_SERVER) || (new->flag & IP_NAT_DYNAMIC)) {
		new->flag |= IP_NAT_BIDIR;
		if (!new->timeout)
			new->timeout = IP_NAT_EXPIRE;
	}
		
#ifdef CONFIG_VIRTUAL_SERVER
	/* additional setup for virtual server rules */
	if (new->flag & IP_NAT_VIRT_SERVER) {
		new->dynamic.vs = kmalloc(sizeof(struct virt_serv), GFP_ATOMIC);
		new->dynamic.vs->sr   = NULL;
		new->dynamic.vs->cl   = NULL;
		new->dynamic.vs->last = NULL;
	}
#endif
	
	cli();
	
	if (!chain) {
		chain = new;
		if (!new->id)
			new->id = 1;
		restore_flags(flags);
		MOD_INC_USE_COUNT;
		return(0);
	}

		
	t = chain;
	last = NULL;  /* pointer to predecessor */
	while(t) {
		if ((t->id >= new->id) && (new->id)) {
			new->next = t;
			if (last)
				last->next = new;
			else
				chain = new;
			if (t->id == new->id) {
				s = t;
				s->id++;
				while ((s) && (s->id == s->next->id)) {
					s = s->next;
					s->id++;
				}
			}
			restore_flags(flags);
			return(0);
		}
		last = t;
		t = t->next;
	}

	if (!new->id)
		new->id = last->id + 1;
	last->next = new;

	restore_flags(flags);
	return(0);
}

#ifdef CONFIG_VIRTUAL_SERVER
int append_to_vs_rule(struct ip_nat *t, struct ip_nat_set *r)
{
	/* function inserts at the head of the chain */
	struct server *new;
	unsigned long flags;

	if (!r->nat.dst) {
		printk("ip_nat (virt. serv.): no server address in virt. server sub-rule\n");
		return(EINVAL);
	}

	for (new=t->dynamic.vs->sr; new; new=new->next)
		if (new->ip == r->nat.dst) {
			printk("ip_nat (virt. serv.): this server already exists\n");
			return(EINVAL);
		}

	new = kmalloc(sizeof(struct server), GFP_ATOMIC);
	new->ip = r->nat.dst;
	new->port = r->nat.ports[0];
	new->pcnt = 0;
	new->bcnt = 0;
	new->connections = 0;
	if (r->weight)
		new->weight = r->weight;
	else
		new->weight = 1;

	save_flags(flags);
	cli();

	new->next = t->dynamic.vs->sr;
	if (!new->next)
		t->dynamic.vs->last = new;
	t->dynamic.vs->sr = new;

	restore_flags(flags);

	return(0);
}

struct server *find_client(struct ip_nat *t,__u32 src_ip)
{
	struct client *cl;

	/* protocol and src_port are unused, they are placeholders for
	 * later, in case somebody implements state information.
	 * so far connections from the same IP get the same server */

	/* horribly inefficient, should use binary tree or hash */
	for (cl = t->dynamic.vs->cl; cl; cl = cl->next) {
		if (cl->ip == src_ip) {
			/* update timer */
			del_timer(&(cl->timer));
			cl->timer.expires = jiffies + t->timeout;
			add_timer(&(cl->timer));
			return(cl->server);
		}
	}

	return(NULL);
}

static void vs_client_expire(unsigned long data)
{
	struct client *cl = (struct client *)data;
        unsigned long flags;

        save_flags(flags);
        cli();

	if (!cl->prev) {
		cl->rule->dynamic.vs->cl = cl->next;
		cl->next->prev = NULL;
	} else {
		cl->prev->next = cl->next;
		cl->next->prev = cl->prev;
	}
	cl->server->connections--;

        restore_flags(flags);

	kfree_s(cl,sizeof(*cl));
}

struct server *new_vs_client(struct ip_nat *t,__u32 src_ip, __u8 protocol, __u16 src_port)
{
	struct client *cl;
	struct server *sr;
	struct server *hsr;
        unsigned long flags, min, hmin;

	/* find server with algorithm specified by the user */
	if (t->flag & IP_NAT_VS_RORO) {
		if (!t->dynamic.vs->last->next)
			t->dynamic.vs->last = t->dynamic.vs->sr;
		else
			t->dynamic.vs->last = t->dynamic.vs->last->next;
		sr = t->dynamic.vs->last;
	} else if (t->flag & IP_NAT_VS_BYTE) {
		min = hmin = t->dynamic.vs->sr->bcnt / t->dynamic.vs->sr->weight;
		sr = t->dynamic.vs->sr;
		for (hsr = t->dynamic.vs->sr->next; hsr; hsr = hsr->next) {
			hmin = hsr->bcnt / hsr->weight;
			if (hmin < min) {
				min = hmin;
				sr = hsr;
			}
		}
	} else if (t->flag & IP_NAT_VS_CONN) {
		sr = NULL;
	} else
		sr = NULL;  /* no valid algorithm specified */

	if (sr)
		sr->connections++;
	else
		return(sr);

	cl = kmalloc(sizeof(struct client), GFP_ATOMIC);
	cl->ip = src_ip;
	cl->port = src_port;
	cl->protocol = protocol;
	cl->prev = NULL;
	cl->rule = t;
	cl->server = sr;

        save_flags(flags);
        cli();
	cl->next = t->dynamic.vs->cl;
	if (cl->next)
		cl->next->prev = cl;
	t->dynamic.vs->cl = cl;
        restore_flags(flags);

	init_timer(&(cl->timer));
	cl->timer.data = (unsigned long)cl;
	cl->timer.function = vs_client_expire;
	cl->timer.expires = jiffies + t->timeout;
	add_timer(&(cl->timer));

	return sr;
}
#endif

#ifdef CONFIG_DYNAMIC_NAT
int append_to_dyn_rule(struct ip_nat *t, struct ip_nat_set *r)
{
	return(0);
}
#endif

#if defined(CONFIG_VIRTUAL_SERVER) || defined(CONFIG_DYNAMIC_NAT)
int append_to_rule (struct ip_nat_set *r) {

	struct ip_nat *t;
	short found = 0;

	/* search the rule where we shall append a sub-rule to */
	t = chain;
	while (t && !found) {
		if (r->nat.id == t->id)
			found = 1;
		if (!found) t = t->next;
	}
	if (!found)
		return(EINVAL);

	if (t->flag & IP_NAT_VIRT_SERVER)
#ifdef CONFIG_VIRTUAL_SERVER
		return(append_to_vs_rule(t,r));
#else
		return(EINVAL);
#endif
	else if (t->flag & IP_NAT_DYNAMIC)
#ifdef CONFIG_DYNAMIC_NAT
		return(append_to_dyn_rule(t,r));
#else
		return(EINVAL);
#endif
	else {
		return(EINVAL);
}
}
#endif

void print_packet (struct iphdr *ip, struct device *dev, struct ip_nat *t)
{
	__u32 *opt = (__u32 *) (ip + 1);
	__u16 src_port=0xFFFF, dst_port=0xFFFF;
	int opti;

        struct tcphdr  *tcp =(struct tcphdr *) ((unsigned long *)ip+ip->ihl);
        struct udphdr  *udp =(struct udphdr *) ((unsigned long *)ip+ip->ihl);
        struct icmphdr *icmp=(struct icmphdr *)((unsigned long *)ip+ip->ihl);

	printk(dev ? dev->name : "-");
	switch(ip->protocol)
	{
		case IPPROTO_TCP:
			printk(" TCP ");
			src_port = ntohs(tcp->source);
			dst_port = ntohs(tcp->dest);
			break;
		case IPPROTO_UDP:
			printk(" UDP ");
			src_port = ntohs(udp->source);
			dst_port = ntohs(udp->dest);
			break;
		case IPPROTO_ICMP:
			printk(" ICMP/%d ", icmp->type);
			break;
		default:
			printk(" PROTO=%d ", ip->protocol);
			break;
	}
	printk("src:");
	print_ip(ip->saddr);
	if(ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP)
		printk(":%hu", src_port);
	printk(" dst:");
	print_ip(ip->daddr);
	if(ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP)
		printk(":%hu", dst_port);
	printk(" nat-src:");
	print_ip(t->src_nat);
	if(t->src_nat_port)
		printk(":%hu", t->src_nat_port);
	printk(" nat-dst:");
	print_ip(t->dst_nat);
	if(t->dst_nat_port)
		printk(":%hu", t->dst_nat_port);
	printk(" L=%hu S=0x%2.2hX I=%hu F=0x%4.4hX T=%hu",
		ntohs(ip->tot_len), ip->tos, ntohs(ip->id),
		ip->frag_off, ip->ttl);
	for (opti=0; 
	     opti<(ip->ihl - sizeof(struct iphdr)/4);
	     opti++)
	printk(" O=0x%8.8X", *opt++);
}

int ip_nat_mod_cntl (int optname, void *tmp_nat, int optlen)
{
	struct ip_nat_set *tmp = (struct ip_nat_set*) tmp_nat;

	if (optlen != sizeof(struct ip_nat_set))
		return(EINVAL);

	switch (tmp->command)
	{
	case IP_NAT_CMD_FLUSH:
		flush_rules();
		break;
	case IP_NAT_CMD_DELETE:
		return (delete_rule(tmp));
		break;
	case IP_NAT_CMD_INSERT:
#ifndef CONFIG_VIRTUAL_SERVER
		if (tmp->nat.flag & IP_NAT_VIRT_SERVER) {
			printk("ip_nat_mod_ctl: no virtual server support compiled in\n");
			return(EINVAL);
		}
#endif
#ifndef CONFIG_DYNAMIC_NAT
		if (tmp->nat.flag & IP_NAT_DYNAMIC) {
			printk("ip_nat_mod_ctl: no dynamic NAT support compiled in\n");
			return(EINVAL);
		}
#endif
		return (insert_rule(tmp));
		break;
	case IP_NAT_CMD_APPEND:
#if defined(CONFIG_VIRTUAL_SERVER) || defined(CONFIG_DYNAMIC_NAT)
		return (append_to_rule(tmp));
#else
		printk("ip_nat_mod_ctl: no virt. server and/or dynamic NAT support\n");
		return (EINVAL);
#endif
		break;
	default:
		printk("ip_nat_mod_ctl: unknown request\n");
		return(EINVAL);
	}
	
		
	return(0);
}

/*
 *  static address translation:
 *  rewrite source- and/or destination ip in the packet
 *  to the new source- and/or destination ip
 */

void ip_nat_translate (struct sk_buff *skb, struct iphdr *ip,
		       struct tcphdr *tcp, struct udphdr *udp,
		       __u32 saddr, __u32 s_msk,
		       __u32 daddr, __u32 d_msk,
		       __u16 sport, __u16 dport)
{
	int len = ntohs(ip->tot_len)-ip->ihl*4;
	
	if (saddr)
		skb->saddr = ip->saddr = (saddr&s_msk) | (ip->saddr&(~s_msk));
	
	if (daddr)
		skb->daddr = ip->daddr = (daddr&d_msk) | (ip->daddr&(~d_msk));

	if (sport) {
		if (ip->protocol == IPPROTO_TCP)
			tcp->source = htons(sport);
		else if (ip->protocol == IPPROTO_UDP)
			udp->source = htons(sport);
	}

	if (dport) {
		if (ip->protocol == IPPROTO_TCP)
			tcp->dest = htons(dport);
		else if (ip->protocol == IPPROTO_UDP)
			udp->dest = htons(dport);
	}

	/*
	 *  that was it, but now we have to recompute
	 *  the checksums
	 */

	switch (ip->protocol) {
		case IPPROTO_TCP:
			tcp->check = 0;
			tcp->check = csum_tcpudp_magic(ip->saddr,
					ip->daddr,len,IPPROTO_TCP,
					csum_partial((char *)tcp,len,0));
			if (tcp->check==0)
				tcp->check = 0xFFFF;
			break;
		case IPPROTO_UDP:
			udp->check = 0;
			udp->check = csum_tcpudp_magic(ip->saddr,
					ip->daddr,len,IPPROTO_UDP,
					csum_partial((char *)udp,len,0));
			if (udp->check==0)
				udp->check = 0xFFFF;
			break;
		default:
			break;
	}
	ip_send_check(ip);

	return;
}

void ip_nat_mod_fnct (u_char direction, struct sk_buff *skb)
{
	struct ip_nat *t;
	struct iphdr *ip;
        struct tcphdr  *tcp;
        struct udphdr  *udp;
        struct icmphdr *icmp;
	__u32 src,dst;
	__u16 src_port=0xFFFF, dst_port=0xFFFF, icmp_type=0xFF;
	char notcpsyn=0, notcpack=0, reverse;
	unsigned short offset;
#if defined(CONFIG_VIRTUAL_SERVER) || defined(CONFIG_DYNAMIC_NAT)
	struct server *tmp = NULL;
#endif

#ifdef KERNEL_2_0
	if ((direction == IP_NAT_IN) && (skb->h.iph))
		ip   = skb->h.iph;
	else if (skb->ip_hdr) /* IP_NAT_OUT */
		ip   = skb->ip_hdr;
	else
		return;  /* no IP packet (yes, we do get here sometimes) */
#else  /* kernel 2.1 */
	ip   = skb->nh.iph;
#endif
        tcp  = (struct tcphdr *) ((unsigned long *)ip+ip->ihl);
        udp  = (struct udphdr *) ((unsigned long *)ip+ip->ihl);
        icmp = (struct icmphdr *)((unsigned long *)ip+ip->ihl);

	if (!chain)
		return;

	if (ip->version != 4)
		return;

	src = ip->saddr;
	dst = ip->daddr;
	offset = ntohs(ip->frag_off) & IP_OFFSET;

	switch(ip->protocol) 
        {
                case IPPROTO_TCP:
                  /* ports stay 0xFFFF if it is not the first fragment */
                        if (!offset) {
                                src_port=ntohs(tcp->source);
                                dst_port=ntohs(tcp->dest);
                                if(!tcp->ack && !tcp->rst)
                                        notcpack=1;
                                if(!tcp->syn || !notcpack)
                                        notcpsyn=1;
                        }
                        break;
                case IPPROTO_UDP:
                  /* ports stay 0xFFFF if it is not the first fragment */
                        if (!offset) {
                                src_port=ntohs(udp->source);
                                dst_port=ntohs(udp->dest);
                        }
                        break;
                case IPPROTO_ICMP:
                  /* icmp_type stays 255 if it is not the first fragment */
                        if (!offset)
                                icmp_type=(__u16)(icmp->type);
                        break;
                default:
                        break;
        }

	for (t=chain; t; t=t->next) 
	{
		/*
		 *  match direction (or option 'bidirectional' is set)
		 */

		if (direction & (t->flag&IP_NAT_DIRECTION_MASK))
			reverse = 0;
		else
			reverse = 1;

		if (reverse && !(t->flag&IP_NAT_BIDIR))
			continue;
		
                /*
                 *  look for a protocol match
                 */
		
		if ((t->protocol) && (t->protocol != ip->protocol))
			continue;

                /*
                 *  look for a VIA address match 
                 */
		
                if ((t->via) && (t->via != skb->dev->pa_addr))
			continue;

                /*
                 *  look for a VIA device match 
                 */
		
                if ((t->via_dev) && (t->via_dev != skb->dev))
			continue;
	
		/*
		 *  match source and destination ip's
		 */

		if (reverse) {
			/* this is more complicated... */
#ifdef CONFIG_VIRTUAL_SERVER
			if (t->flag & IP_NAT_VIRT_SERVER) {
				/*  if !(src-ip is one of the real servers) continue */
				for (tmp = t->dynamic.vs->sr; tmp; tmp = tmp->next) {
					if (tmp->ip == src)
						break;
				}  /* value of tmp is also used below */
				if (!tmp)
					continue;

				if ((t->src_nat) && ((dst&t->src_msk) != t->src))
						continue;
			} else {
#endif
#ifdef CONFIG_DYNAMIC_NAT
#endif
				if (t->dst_nat) {
					if ((src&t->dst_nat_msk) != t->dst_nat)
						continue;
				} else {
					if ((src&t->dst_msk) != t->dst)
						continue;
				}
				if (t->src_nat) {
					if ((dst&t->src_nat_msk) !=
					     t->src_nat)
					continue;
				} else {
					if ((dst&t->src_msk) != t->src)
						continue;
				}
#if defined(CONFIG_VIRTUAL_SERVER) || defined(CONFIG_DYNAMIC_NAT)
			}
#endif
		} else {
			/* ...than this case */
			if (!((src&t->src_msk)==t->src &&
			      (dst&t->dst_msk)==t->dst))
				continue;
		}

		/*
		 *  match ports
		 */

		if (t->nsp + t->ndp) {
			if (offset)
				printk("IP NAT-Warning: found fragment, ignoring ports\n");
			if ((t->flag & IP_NAT_VIRT_SERVER) && (reverse) &&
			    (ip->protocol==IPPROTO_TCP || ip->protocol==IPPROTO_UDP)) {
#ifdef CONFIG_VIRTUAL_SERVER
				if ((tmp->port) && (t->ndp == 1)) {
					/* port has been remapped */
					if (!port_match(&tmp->port, 1, src_port, 0))
						continue;
				} else {
					/* port has not been remapped */
					if (!port_match(&t->ports[t->nsp], t->ndp, src_port,
						       t->flag&IP_NAT_DRNG))
						continue;
				}
#endif
			} else if((ip->protocol==IPPROTO_ICMP &&
				!port_match(&t->ports[0], t->nsp,
					    icmp_type,t->flag&IP_NAT_SRNG)) ||
				!(ip->protocol==IPPROTO_ICMP || ((!reverse) &&
				port_match(&t->ports[0], t->nsp, src_port,
					   t->flag&IP_NAT_SRNG) &&
				port_match(&t->ports[t->nsp], t->ndp, dst_port,
					   t->flag&IP_NAT_DRNG)) || ((reverse) &&
				port_match(t->src_nat_port ? &(t->src_nat_port) : &t->ports[0], 
					   t->src_nat_port ? 1 : t->nsp,
					   dst_port, 
					   t->src_nat_port ? 0 : t->flag&IP_NAT_SRNG) &&
				port_match(t->dst_nat_port ? &(t->dst_nat_port) : &t->ports[t->nsp],
					   t->dst_nat_port ? 1 : t->ndp,
					   src_port,
					   t->dst_nat_port ? 0 : t->flag&IP_NAT_DRNG))))
			{
				continue;
			}
		}


		/* if we get here, we've found a matching rule */

		/*
		 *  print matching packets
		 */

		if (t->flag & IP_NAT_LOG) {
			if (direction == IP_NAT_IN)
				printk("IP NAT in (before) ");
			else
				printk("IP NAT out (before) ");
			print_packet(ip, skb->dev, t);
			printk("\n");
		}

		/*
		 *  process packet according to rule
		 */

		/* increment packet and byte counters */
		t->pcnt++;
		t->bcnt+=ntohs(ip->tot_len);

		/* do-nothing rule */
		if (t->flag & IP_NAT_SKIP)
			return;

		if (t->flag & IP_NAT_DYNAMIC) {
#ifdef CONFIG_DYNAMIC_NAT
			/* not implemented yet */
#endif
		} else if (t->flag & IP_NAT_VIRT_SERVER) {
#ifdef CONFIG_VIRTUAL_SERVER

			/*
			 * Address Translation for Load Balancing
			 * (Virtual Servers)
			 */
			
			if (!t->dynamic.vs->sr)
				return;

			if (reverse) {
				for (tmp = t->dynamic.vs->sr; tmp; tmp = tmp->next)
					if (tmp->ip == ip->saddr) {
						tmp->pcnt++;
						tmp->bcnt+=ntohs(ip->tot_len);
						break;
					}
				ip_nat_translate(skb, ip, tcp, udp,
						 t->dst,        /* virtual server IP */
						 0xffffffff,
					         t->src,
						 t->src_msk,
						 (t->ndp == 1) ? t->ports[t->nsp] : 0,
						 0);
			} else {
				if ((ip->protocol == IPPROTO_TCP) &&
				    (!(tmp = find_client(t,ip->saddr))))
					tmp = new_vs_client(t,ip->saddr,ip->protocol,tcp->source);
				else if (!(tmp = find_client(t,ip->saddr)))
						tmp = new_vs_client(t,ip->saddr,ip->protocol,udp->source);

				if (!tmp)
					return;

				tmp->pcnt++;
				tmp->bcnt+=ntohs(ip->tot_len);
				ip_nat_translate(skb, ip, tcp, udp,
						 t->src_nat,
						 t->src_nat_msk,
					         tmp->ip,      /* IP of the real server */
						 0xffffffff,
						 0,
						 /* port rewriting is only possible when it is
						  * clear what port to insert on the way back.
						  * remember, we don't keep state info */
						 (t->ndp == 1) ? tmp->port : 0);
			}
#endif
		} else {

			/*
			 * (classic) Static Address Translation
			 */
			
			if (reverse) {
				/* reverse match (bidirectional rule) */
				ip_nat_translate(skb, ip, tcp, udp,
						 t->dst,
						 t->dst_msk,
					         t->src,
						 t->src_msk,
						 t->dst_nat_port,
						 t->src_nat_port);
			} else {
				/* direct match */
				ip_nat_translate(skb, ip, tcp, udp,
						 t->src_nat,
						 t->src_nat_msk,
					         t->dst_nat,
						 t->dst_nat_msk,
						 t->src_nat_port,
						 t->dst_nat_port);
			}
		}

		/* change route-address only if target host is local */
		if (dst == skb->raddr)
			skb->raddr = ip->daddr;
		
		/*
		 *  print matching packets
		 */

		if (t->flag & IP_NAT_LOG) {
			if (direction == IP_NAT_IN)
				printk("IP NAT in (after) ");
			else
				printk("IP NAT out (after) ");
			print_packet(ip, skb->dev, t);
			printk("\n");
		}


		return;
		
	}  /* loop */
}

int register_ip_nat_app (void)
{
	return(0);
}

int unregister_ip_nat_app (void)
{
	return(0);
}



int init_module (void)
{
#ifdef KERNEL_2_0
	register_symtab (&ip_nat_module_syms);
#endif

	/* tell the kernel the addresses of these two functions */
        if ( register_nat(ip_nat_mod_fnct, ip_nat_mod_cntl) != 0 )
                return(-EIO);

#ifdef CONFIG_PROC_FS
        proc_net_register(&(struct proc_dir_entry) {
                PROC_NET_IPNAT, 6, "ip_nat",
                S_IFREG | S_IRUGO, 1, 0, 0,
                0, &proc_net_inode_operations,
                ip_nat_procinfo
        });
#ifdef CONFIG_VIRTUAL_SERVER
        proc_net_register(&(struct proc_dir_entry) {
                PROC_NET_IPNAT_VSERVER, 18, "ip_nat_virt_server",
                S_IFREG | S_IRUGO, 1, 0, 0,
                0, &proc_net_inode_operations,
                ip_nat_vs_procinfo
        });
#endif /* CONFIG_VIRTUAL_SERVER */
#endif /* CONFIG_PROC_FS */

        return(0);
}

void cleanup_module (void)
{
#ifdef CONFIG_PROC_FS
	proc_net_unregister(PROC_NET_IPNAT);
#ifdef CONFIG_VIRTUAL_SERVER
	proc_net_unregister(PROC_NET_IPNAT_VSERVER);
#endif /* CONFIG_VIRTUAL_SERVER */
#endif /* CONFIG_PROC_FS */
	flush_rules();

	unregister_nat();
}
