/*
 *	IP Dynamic NAT implementation by Michael Hasenstein
 *		mha@informatik.tu-chemnitz.de
 *		mha@debis-sfi.de
 *      (see RFC 1631)
 *
 *
 *      Copyright (c) 1997 by Michael Hasenstein
 *
 *      --------------------------------------------------------------------
 *      parts of this code, especially the packet matching stuff, have been
 *      taken from linux/net/ipv4/ip_fw.c
 *      see this file for copyrights for this code
 *      --------------------------------------------------------------------
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef MODULE
#error This file must be compiled as a module.
#endif

#include <linux/config.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ip.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <net/icmp.h>
#include "ip_nat_module.h"

#ifdef CONFIG_PROC_FS
static int ip_dynamic_nat_procinfo (char *buffer, char **start, off_t offset,
                                    int length, int reset)
{
	off_t /*pos=0,*/ begin=0;
        struct ip_nat *i;
        unsigned long flags;
        int len/*, p*/;
        /*int last_len = 0;*/
	
	i = chain;

	len=sprintf(buffer, "IP DYNAMIC NAT rules\n");

	save_flags(flags);
        cli();

	/* output */
	
	restore_flags(flags);
        *start=buffer+(offset-begin);
        len-=(offset-begin);
        if(len>length)
                len=length;     
	
        return(len);
}
#endif  /* CONFIG_PROC_FS */

void ip_nat_dynamic (struct sk_buff *skb, struct iphdr *ip,
		       struct tcphdr *tcp, struct udphdr *udp,
		       __u32 saddr, __u32 s_msk,
		       __u32 daddr, __u32 d_msk)
{
	return;
}


int init_module (void)
{
#ifdef KERNEL_2_0
	register_symtab (&ip_nat_module_syms);
#endif

#ifdef CONFIG_PROC_FS
        proc_net_register(&(struct proc_dir_entry) {
                PROC_NET_IPNAT_DYNAMIC, 14, "ip_nat_dynamic",
                S_IFREG | S_IRUGO, 1, 0, 0,
                0, &proc_net_inode_operations,
                ip_dynamic_nat_procinfo
        });
#endif

        return(0);
}

void cleanup_module (void)
{
#ifdef CONFIG_PROC_FS
	proc_net_unregister(PROC_NET_IPNAT_DYNAMIC);
#endif
	flush_dynamic_rules();
}
