/*
 *      IP nat module functionality definitions
 */

#ifndef _IP_NAT_MODULE_H
#define _IP_NAT_MODULE_H

#include <asm/errno.h>
#include <asm/byteorder.h>
#include <net/ip_nat.h>
#include <netinet/in.h>

/* defines for the flag field in struct ip_nat */
#define IP_NAT_DIRECTION_MASK 0x03
#define IP_NAT_IN       0x0001  /* equals IN_NAT in inlcude/net/ip_nat.h */
#define IP_NAT_OUT      0x0002  /* equals OUT_NAT */
#define IP_NAT_BIDIR    0x0004
#define IP_NAT_SKIP     0x0008
#define IP_NAT_LOG      0x0010
#define IP_NAT_SRNG     0x0020
#define IP_NAT_DRNG     0x0040
#define IP_NAT_DYNAMIC  0x0080
 

#define IP_NAT_CMD_FLUSH     0x01
#define IP_NAT_CMD_INSERT    0x02
#define IP_NAT_CMD_DELETE    0x04
#define IP_NAT_CMD_TIMEOUTS  0x08


#define print_ip(a)   printk("%ld.%ld.%ld.%ld",     \
                              (ntohl(a)>>24) & 0xFF,\
                              (ntohl(a)>>16) & 0xFF,\
                              (ntohl(a)>>8)  & 0xFF,\
                              (ntohl(a))     & 0xFF);


#define IP_NAT_MAX_PORTS 10
struct ip_nat {
	struct ip_nat  *next;                    /* pointer to next rule */
	void           *dynamic;                 /* pointer to dynamic rules */
	int            id;                       /* unique number for this rule */
        struct in_addr src, dst;                 /* source/destination ip */
        struct in_addr src_msk, dst_msk;         /* network masks */
        unsigned short nsp, ndp;                 /* number of source/destination ports */
        __u16          ports[IP_NAT_MAX_PORTS];  /* array of port numbers */
	__u8           protocol;                 /* protocol id (udp,tcp,icmp) */
        struct in_addr via;                      /* interface ip */
	struct device *via_dev  ;                /* device of interface "via" */
        struct in_addr src_nat, dst_nat;         /* new source/destination ip */
        struct in_addr src_nat_msk, dst_nat_msk; /* network masks */
	__u16          src_nat_port, dst_nat_port; /* new source/destination port */
	unsigned long  pcnt,bcnt;                /* Packet and byte counters */
        unsigned short flag;                     /* flag: bidirectional, skip, packet logging, \
	                                                  direction (in/out), dynamic rule */
};


/* for calls to setsockopt() (e.g. ipnatadm) */
struct ip_nat_set {
	struct ip_nat  nat;
	int command;
	int timeout;
        char vianame[IFNAMSIZ];        /* interface name */
};


extern int register_ip_nat_app(void);
extern int unregister_ip_nat_app(void);

#endif /* _IP_NAT_MODULE_H */
